/*
 * Copyright (c) 2015 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Tue Jun  2 2015
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <poll.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <netinet/tcp.h>

#include "libgenomix.h"


/* --- local data ---------------------------------------------------------- */

static struct connections {
  const char *host;
  const char *session;
  int cmdfd, evtfd;

  struct http_rqst rqst;
  struct http_event evt;
} *connections;

static uint8_t nconnection;
static uint8_t connectionid;

static int	genomix_socket(const char *host, const char *port, char *emsg,
			size_t emsgsize);
static int	genomix_http_buffer(struct http_buffer *buffer);


/* --- genomix_connect ----------------------------------------------------- */

int
genomix_connect(char *hostport, char *reqsession, char *emsg, size_t emsgsize)
{
  struct connections conn;
  char *host, *port;
  char uri[256];
  float v;
  int s;

  if (hostport) {
    host = hostport;
    if ((port = strchr(host, ':'))) {
      *port++ = '\0';
    } else
      port = "8080";
  } else {
    host = "localhost";
    port = "8080";
  }
  conn.host = strdup(host);
  conn.session = NULL;
  conn.cmdfd = conn.evtfd = -1;
  if (genomix_http_buffer(&conn.rqst.uri)) goto err;
  if (genomix_http_buffer(&conn.rqst.buffer)) goto err;
  if (genomix_http_buffer(&conn.rqst.args)) goto err;
  if (genomix_http_buffer(&conn.evt.buffer)) goto err;

  /* connect to genomix */
  conn.cmdfd = genomix_socket(host, port, emsg, emsgsize);
  if (conn.cmdfd < 0) goto err;

  s = http_transaction(conn.cmdfd, &conn.rqst, "get", "/version", NULL);
  if (s < 200 || s >= 300) {
    snprintf(emsg, emsgsize, "unable to get genomixd server version: %s",
             conn.rqst.body);
    goto err;
  }
  if (sscanf(conn.rqst.body, "%f", &v) != 1) {
    snprintf(emsg, emsgsize, "not a genomixd server");
    goto err;
  }
  if (v < 1.5) {
    snprintf(emsg, emsgsize, "unsupported genomixd server version: %s",
             conn.rqst.body);
    goto err;
  }


  /* request a session */
  snprintf(uri, sizeof(uri), "/session/start?s=%s", reqsession?reqsession:"");
  s = http_transaction(conn.cmdfd, &conn.rqst, "get", uri, NULL);
  if (s < 200 || s >= 300) {
    snprintf(emsg, emsgsize, "%s", conn.rqst.body);
    goto err;
  }
  conn.session = strdup(conn.rqst.body);


  /* event socket */
  conn.evtfd = genomix_socket(host, port, emsg, emsgsize);
  if (conn.evtfd < 0) return -1;

  snprintf(uri, sizeof(uri), "/session/listen?s=%s", conn.session);
  s = http_transaction(conn.evtfd, &conn.rqst, "get", uri, NULL);
  if (s < 200 || s >= 300) {
    snprintf(emsg, emsgsize, "%s", conn.rqst.body);
    goto err;
  }


  /* register connection */
  s = connectionid++;
  if (s >= nconnection) {
    struct connections *c = realloc(connections, (s+1)*sizeof(*c));
    if (!c) goto err;
    connections = c;
    nconnection = s+1;
  }
  connections[s] = conn;

  return s;
err:
  if (conn.host) free((void *)conn.host);
  if (conn.session) free((void *)conn.session);
  if (conn.cmdfd >= 0) close(conn.cmdfd);
  if (conn.evtfd >= 0) close(conn.evtfd);
  if (conn.rqst.uri.data) free(conn.rqst.uri.data);
  if (conn.rqst.buffer.data) free(conn.rqst.buffer.data);
  if (conn.rqst.args.data) free(conn.rqst.args.data);
  if (conn.evt.buffer.data) free(conn.evt.buffer.data);
  return -1;
}


/* --- genomix_disconnect -------------------------------------------------- */

int
genomix_disconnect(int id)
{
  struct connections *conn;
  int i, remain;

  for (i=remain = 0; i<nconnection; i++)
    if (connections[i].cmdfd >= 0) remain++;

  if (id >= nconnection) return remain;
  conn = connections + id;

  if (conn->host) free((void *)conn->host);
  conn->host = NULL;
  if (conn->session) free((void *)conn->session);
  conn->session = NULL;
  if (conn->evtfd >= 0) close(conn->evtfd);
  conn->evtfd = -1;
  if (conn->cmdfd >= 0) close(conn->cmdfd);
  conn->cmdfd = -1;
  if (conn->rqst.uri.data) free(conn->rqst.uri.data);
  conn->rqst.uri.data = NULL;
  if (conn->rqst.buffer.data) free(conn->rqst.buffer.data);
  conn->rqst.buffer.data = NULL;
  if (conn->rqst.args.data) free(conn->rqst.args.data);
  conn->rqst.args.data = NULL;
  if (conn->evt.buffer.data) free(conn->evt.buffer.data);
  conn->evt.buffer.data = NULL;

  return remain - 1;
}


/* --- genomix_get --------------------------------------------------------- */

const char *
genomix_get(int id, const char *cmd, const char *args[],
            char *emsg, size_t emsgsize)
{
  struct connections *conn;

  conn = connections + id;
  if (id >= nconnection || conn->cmdfd < 0) {
    snprintf(emsg, emsgsize, "no such connection %d", id);
    return NULL;
  }

  if (http_get(conn->cmdfd, &conn->rqst, conn->session, cmd, args)) {
    snprintf(emsg, emsgsize, "%s", conn->rqst.body);
    return NULL;
  }

  return conn->rqst.body;
}


/* --- genomix_doevent ----------------------------------------------------- */

void
genomix_doevent(int timeout)
{
  struct pollfd fds[nconnection];
  int i, s;

  for (i=0; i<nconnection; i++) {
    fds[i].fd = connections[i].evtfd;
    fds[i].events = POLLIN;
  }

  s = poll(fds, nconnection, timeout);
  if (s <= 0) return;

  for (i=0; i<nconnection; i++)
    if (fds[i].revents & POLLIN) {
      struct connections *c = &connections[i];
      do {
        if (http_event(c->evtfd, &c->evt)) break;
        if (!strcmp(c->evt.event, "request"))
          genomix_event(c->evt.data);
        else if (!strcmp(c->evt.event, "quit")) {
          genomix_disconnect(i);
          break;
        }
      } while(c->evt.buffer.end > c->evt.buffer.start);
    }
}


/* --- genomix_socket ------------------------------------------------------ */

static int
genomix_socket(const char *host, const char *port, char *emsg, size_t emsgsize)
{
  struct addrinfo hints, *res, *res0;
  const char *cause;
  int fd, f;
  int e, ecause;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  e = getaddrinfo(host, port, &hints, &res0);
  if (e) {
    snprintf(emsg, emsgsize, "%s", gai_strerror(e));
    return -1;
  }

  fd = -1;
  for (res = res0; res; res = res->ai_next) {
    if (res->ai_family != AF_INET && res->ai_family != AF_INET6) continue;

    fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (fd < 0) { cause = "socket"; ecause = errno; continue; }

    if (connect(fd, res->ai_addr, res->ai_addrlen) < 0) {
      cause = "connect"; ecause = errno;
      close(fd);
      fd = -1;
      continue;
    }

    setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (char *)&(int){1}, sizeof(int));
    break;
  }
  freeaddrinfo(res0);

  if (fd < 0) {
    snprintf(emsg, emsgsize, "%s: %s", cause, strerror(ecause));
    return -1;
  }

  f = fcntl(fd, F_GETFL);
  if (f != -1) {
    f |= O_NONBLOCK;
    fcntl(fd, F_SETFL, f);
  }

  return fd;
}


/* --- genomix_http_buffer ------------------------------------------------- */

static int
genomix_http_buffer(struct http_buffer *buffer)
{
  buffer->data = malloc(64);
  if (!buffer->data) return -1;

  buffer->len = 64;
  buffer->end = buffer->data;
  buffer->start = buffer->data;
  return 0;
}
