%

%
% Copyright (c) 2015 LAAS/CNRS
% All rights reserved.
%
% Redistribution  and  use  in  source  and binary  forms,  with  or  without
% modification, are permitted provided that the following conditions are met:
%
%   1. Redistributions of  source  code must retain the  above copyright
%      notice and this list of conditions.
%   2. Redistributions in binary form must reproduce the above copyright
%      notice and  this list of  conditions in the  documentation and/or
%      other materials provided with the distribution.
%
% THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
% WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
% MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
% ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
% WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
% ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
% IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
%
%                                           Anthony Mallet on Tue Jun  2 2015
%
classdef client < handle
  properties (Access = private)
    host
    session
    id
  end

  methods
    function self = client(host, session)
      % genomix.client - Create connections to a genomix server.
      %
      % c = genomix.client()
      %    Connect to a server on 'localhost' port 8080.
      % c = genomix.client('host:port')
      %    Connect to a server on 'host' and 'port'.
      %
      % genomix.client Methods:
      %    rpath - Add a new load path on the server.
      %    load - Load a component on the server.
      %    update - Process received events.
      %
      switch nargin
        case 0
          host = 'localhost:8080';
          session = '';
        case 1
          session = '';
      end

      self.host = host;
      self.session = session;
      self.id = genomix_server('connect', self.host, self.session);
    end

    function delete(self)
      % delete - Disconnect an instance of a genomix.client.
      %
      genomix_server('disconnect', self.id);
    end


    % --- rpath ------------------------------------------------------------
    %
    function r = rpath(self, rpath)
      % rpath - Add a new load path on the server.
      %
      r = genomix_server('get', self.id, sprintf('/rpath/%s', rpath));
    end


    % --- load -------------------------------------------------------------
    %
    function r = load(self, client, varargin)
      % load - Load a component on the server.
      %
      % c = client.load(component) -
      %   Load 'component' in the server a return the corresponding
      %   genomix.component handle.
      %
      % c = client.load(component, '-h') -
      %   Return the available load options for the 'component'.
      %
      usage = 0;
      argv = {};
      for i = 1:length(varargin)
        if varargin{i} == '-h'
          usage = 1;
        else
          argv = { argv{:} varargin{i} };
        end
      end
      argv = [ '{' strjoin(argv) '}' ];

      if usage
        r = sprintf([ ...
           'Usage:\n\t client.load(''%s'', [-h], [client options])\n\n' ...
           'Client options:\n' ], client);
        opts = genomix_server('get', self.id, strcat('/usage/', client));
        for i = 1:length(opts)
            r = sprintf('%s  %-26s %s\n', r, opts{i}.option, opts{i}.descr);
        end
        return
      end

      module = genomix_server('get', self.id, ['/load/' client], 'argv', argv);
      meta = genomix_server('get', self.id,  ['/module/' module '/info']);
      r = genomix.component(self, client, meta);
    end


    % --- reconnect --------------------------------------------------------
    %
    function reconnect(self)
      try
        genomix_server('disconnect', self.id);
      catch
      end
      self.id = genomix_server('connect', self.host, self.session);
    end


    % --- update -----------------------------------------------------------
    %
    function update(self)
      genomix_server('doevents', 0);
    end


    % --- rqst -------------------------------------------------------------
    %
    function id = rqst(self, client, name, meta, defarg, varargin)

      % process options
      fmt = 'dict';
      wait = 'done';
      timeout = -1;

      while ~isempty(varargin)
        if ~isstr(varargin{1}) break; end
        arg = varargin{1}; varargin(1) = [];
        switch arg
          case { '-a', '-ack' }
            wait = 'sent';
          case { '-t', '-timeout' }
            timeout = varargin{1};
            varargin(1) = [];
          case { '-f', '-flat' }
            fmt = 'flat';
          case '-args';
            id = meta;
            return;

          case '--'
            break;
          case '-h'
            id = sprintf([ 'Usage:\n\t %s.%s ' ...
                           '[-a|-ack] [-t|-timeout secs] [-f|-flat]\n\t\t' ...
                           '[-args] [-h] [--] input [callback]\n\n' ...
                           'Input dictionary:%s' ], client, name, ...
                         self.meta2doc(meta.input, meta.meta));
            return

          otherwise
            if arg(1) == '-'
              error(['unknown option -- ' arg]);
            end
            varargin = { arg varargin{:} };
            break
        end
      end

      % look for callback
      callback = [];
      if ~isempty(varargin) && isa(varargin{end}, 'function_handle')
         callback = varargin{end};
         varargin(end) = [];
         wait = 'sent';
      end

      % create json argument
      switch length(varargin)
        case 0
          if (fmt == 'flat') error('-f requires arguments'); end
          varargin = self.iscanargs(meta.input, meta.meta);

        case 1
          if strcmp(fmt, 'flat') || ~isstruct(varargin{1})
            fmt = 'flat';
          else
            varargin = varargin{1};
          end

        otherwise
          fmt = 'flat';
      end
      json = self.arg2json(fmt, varargin, defarg, meta.input);

      % send remotely
      i = genomix_server('get', self.id, ...
                  ['/module/' client '/send/' name], 'input', json);

      % register request
      id = genomix.request(self.id, client, i);
      r = struct();
      r.rh = id;
      r.callback = callback;
      r.status = 'none';
      rqstids = genomix.client.rqstids;
      if ~mislocked mlock; end
      rqstids(sprintf('%s::%s', client, i)) = r;

      % wait for completion
      id.wait(timeout, wait);
    end


    % --- port -------------------------------------------------------------
    %
    function data = port(self, client, name, meta, varargin)
      nvarargin = length(varargin);
      if meta.simple & nvarargin ~= 0
        error(sprintf('%s::%s expects no argument', client, name));
      elseif ~meta.simple
        if nvarargin ~= 1
           error(sprintf('%s::%s expects a port handle name argument', ...
                         client, name));
        end
        name = [name '/' varargin{1}];
      end

      data = genomix_server('get', self.id, ['/module/' client '/read/' name]);
    end

  end

  methods (Static)
    function doevents(data)
      if ~isfield(data, 'id') return; end

      rqstids = genomix.client.rqstids;
      if rqstids.isKey(data.id)
        r = rqstids(data.id);
        for p = fieldnames(data)'
          r.(p{1}) = data.(p{1});
          switch p{1}
            case 'status'
              r.rh.status = data.status;
            case 'output'
              r.rh.result = data.output;
            case 'errmsg'
              r.rh.exception = data.errmsg;
          end
        end

        if ~isempty(r.callback)
          cb = r.callback;
          cb(r);
        end

        if isfield(data, 'status')
          if strcmp(data.status, 'done') || strcmp(data.status, 'error')
            remove(rqstids, data.id);
            if isempty(rqstids.keys) munlock; end
          end
        end
      end
    end

    function val = rqstids()
      persistent pval;
      if isempty(pval)
        pval = containers.Map;
      end
      val = pval;
    end
  end

  methods (Access = private)

    % --- iscanarg ---------------------------------------------------------
    %
    function input = iscanarg(self, kind, prompt, defval)
      if ~isempty(defval)
        if isnumeric(defval)
          prompt = sprintf('%s (%g)', prompt, defval);
        else
          prompt = sprintf('%s (%s)', prompt, defval);
        end
      end

      while 1
        p = sprintf('input(''%s > '')', prompt);
        input = evalin('base', p);

        % user might want default value or interrupt
        if isempty(input) input = defval; end
        if isstr(input) && strcmp(input, '...')
          input = [];
          return
        end

        % check input validity
        try
          switch kind{1}
            case 'boolean'
              if input input = true; else input = false; end
            case 'octet'
              if ~isnumeric(input)
                error(sprintf('expecting %s', kind{1}));
              end
              if input < -128 || input > 127
                error(sprintf('argument out of range, expecting %s', kind{1}));
              end
            case 'short'
              if ~isnumeric(input)
                error(sprintf('expecting %s', kind{1}));
              end
              if input < -32768 || input > 32767
                error(sprintf('argument out of range, expecting %s', kind{1}));
              end
            case 'unsigned short'
              if ~isnumeric(input)
                error(sprintf('expecting %s', kind{1}));
              end
              if input < 0 || input > 65535
                error(sprintf('argument out of range, expecting %s', kind{1}));
              end
            case 'long'
              if ~isnumeric(input)
                error(sprintf('expecting %s', kind{1}));
              end
              if input < -2^31 || input > 2^31-1
                error(sprintf('argument out of range, expecting %s', kind{1}));
              end
            case 'unsigned long'
              if ~isnumeric(input)
                error(sprintf('expecting %s', kind{1}));
              end
              if input < 0 || input > 2^32-1
                error(sprintf('argument out of range, expecting %s', kind{1}));
              end
            case 'long long'
              if ~isnumeric(input)
                error(sprintf('expecting %s', kind{1}));
              end
              if input < -2^63 || input > 2^63-1
                error(sprintf('argument out of range, expecting %s', kind{1}));
              end
            case 'unsigned long long'
              if ~isnumeric(input)
                error(sprintf('expecting %s', kind{1}));
              end
              if input < 0 || input > 2^64-1
                error(sprintf('argument out of range, expecting %s', kind{1}));
              end
            case {'float' 'double'}
              if ~isnumeric(input)
                error(sprintf('expecting %s', kind{1}));
              end
            case 'char'
              if ~isstr(input)
                error(sprintf('expecting %s', kind{1}));
              end
              input = input{1};
            case 'string'
              if ~isstr(input)
                error(sprintf('expecting %s', kind{1}));
              end
              if length(kind) > 1
                 if length(input) > kind{2}
                   error(sprintf('string too long'));
                 end
              end
            case 'enum'
              if find(ismember(kind{2}, input))
              else
                error(sprintf('expecting %s', strjoin(kind{2}, ', ')));
              end
          end
        catch e
          disp(e.message);
          continue;
        end

        break;
      end
    end


    % --- iscandict --------------------------------------------------------
    %
    function arg = iscandict(self, indent, name, format, meta)
      prompt = sprintf('%s %s %s', indent, format.kind, name);
      if isfield(format, 'length')
        prompt = sprintf('%s[%d]', prompt, format.length);
      end
      if isfield(meta, 'doc')
        prompt = sprintf('%s: %s', prompt, meta.doc);
      end
      if isfield(meta, 'items') items = meta.items; else items = struct(); end
      if isfield(meta, 'default')
        default = meta.default;
      else
        default = [];
      end

      switch format.kind
        case {'struct' 'union'}
          disp(prompt);
          arg = struct();
          for p = fieldnames(format.type)'
            if isfield(items, p{1}) m = items.(p{1}); else m = struct(); end

            a = self.iscandict(sprintf('%s| ', indent), ...
                               sprintf('%s.%s', name, p{1}), ...
                               format.type.(p{1}), m);
            if isempty(a) break; end
            arg.(p{1}) = a;
          end

        case {'array' 'sequence'}
          disp(prompt);
          if isfield(format, 'length')
            len = format.length;
          else
            len = self.iscanarg({'unsigned long'}, ...
                                sprintf('%s length', prompt), 0);
            if isempty(len) arg = {}; return; end
          end

          arg = cell(len, 1);
          for i = 1:len
            if iscell(items) && length(items) >= i
              m = items{i};
            else
              m = struct();
            end

            a = self.iscandict(sprintf('%s| ', indent), ...
                               sprintf('%s[%d]', name, i), ...
                               format.type, m);
            if isempty(a) break; end
            arg(i) = { a };
          end

        case 'optional'
          disp(prompt);
          pres = self.iscanarg({'boolean'}, ...
                               sprintf('%s| present', indent), 1);
          if isempty(pres) || ~pres arg = {}; return; end
          m = struct('items', items, 'default', default);
          a = self.iscandict(sprintf('%s| ', indent), name, ...
                               format.type, m);
          if isempty(a) arg = {}; return; end
          arg = {a};

        otherwise
          kind = { format.kind };
          if isfield(format, 'length')
            kind(end+1) = { format.length };
          end
          if isfield(format, 'values')
            kind = { kind{:} format.values{:} };
          end
          arg = self.iscanarg(kind, prompt, default);
      end
    end


    % --- iscanargs --------------------------------------------------------
    %
    function arg = iscanargs(self, input, meta)
      arg = struct();
      for p = fieldnames(input)'
        if isfield(meta, p{1}) m = meta.(p{1}); else m = struct(); end
        a = self.iscandict('', p{1}, input.(p{1}), m);
        if isempty(a) break; end
        arg.(p{1}) = a;
      end
    end


    % --- dict2json --------------------------------------------------------
    %
    function json = dict2json(self, arg, data, format)
      switch format.kind
        case {'struct', 'union'}
          % check missing arguments
          if ~isstruct(data)
             error(sprintf('%s %s expects structure data', format.kind, arg));
          end
          for p = fieldnames(format.type)'
            if ~isfield(data, p{1})
              error(sprintf('missing value for argument %s.%s', arg, p{1}));
            end
          end

          % build json
          json = {};
          for p = fieldnames(data)'
            if ~isfield(format.type, p{1})
              error(sprintf('unknonwn argument %s.%s', arg, p{1}));
            end
            v = self.dict2json(sprintf('%s.%s', arg, p{1}), ...
                               data.(p{1}), format.type.(p{1}));
            json(end+1) = { sprintf('"%s":%s', p{1}, v) };
          end
          json = ['{' strjoin(json, ',') '}'];

        case {'array', 'sequence'}
          % check missing arguments
          if ~iscell(data)
             error(sprintf('%s %s expects cell array data', format.kind, arg));
          end
          if isfield(format, 'length')
            l = format.length;
            if length(data) ~= l
              error(sprintf('%s %s has wrong size: must be %d elements', ...
                            format.kind, arg, l));
            end
          else
            l = length(data);
          end
          for i = 1:l
            if isempty(data{i})
              error(sprintf('missing value for index %d of argument %s', ...
                            i, arg));
            end
          end

          % build json
          json = cell(l, 1);
          for i = 1:l
            json(i) = { self.dict2json(sprintf('%s[%d]', arg, i), ...
                                       data{i}, format.type) };
          end
          json = ['[' strjoin(json, ',') ']'];

        case 'optional'
          if isempty(data)
            json = 'null';
          else
            json = self.dict2json(arg, data, format.type);
          end

        case {'char', 'string'}
          if isfield(format, 'length')
            if length(data) > format.length
              error(sprintf('%s %s too long: %d max', format.kind, arg, l));
            end
          end
          json = strrep(data, '\', '\\');
          json = strrep(data, '"', '\"');
          json = ['"' data '"'];

        case 'enum'
          if find(ismember(format.values, data))
            json = sprintf('"%s"', data);
          else
            error(sprintf('bad value for %s %s: %s (must be one of %s)', ...
                          format.kind, arg, data, ...
                          strjoin(format.values, ', ')));
          end

        otherwise
          if ~isnumeric(data)
            error(sprintf('bad numeric value for %s %s: %s', ...
                          format.kind, arg, data));
          end
          json = num2str(data);
      end
    end


    % --- arg2json ---------------------------------------------------------
    %
    function json = arg2json(self, fmt, data, defarg, format)
      % create argument dictionary if needed
      if fmt == 'flat'
        args = data;
        data = struct();
        for p = fieldnames(format)'
          if isempty(args) break; end
          [data.(p{1}), args] = self.args2dict(format.(p{1}), args);
        end

        if (length(args))
          error(sprintf('too many arguments: %s', strjoin(args))); end
      end

      % merge defaults
      for k = fieldnames(defarg)'
        if ~isfield(data, k{1})
          data.(k{1}) = defarg.(k{1});
        end
      end

      % check missing arguments
      for p = fieldnames(format)'
        if ~isfield(data, p{1})
          error(sprintf('missing value for argument %s', p{1}));
        end
      end

      % build json string
      json = {};
      for p = fieldnames(data)'
        if ~isfield(format, p{1})
          error(sprintf('unknown argument %s', p{1}));
        end
        v = self.dict2json(p{1}, data.(p{1}), format.(p{1}));
        json(end+1) = { sprintf('"%s":%s', p{1}, v) };
      end
      json = ['{' strjoin(json, ',') '}'];
    end


    % --- args2dict ---------------------------------------------------------
    %
    function [arg, args] = args2dict(self, format, args)
      switch format.kind
        case {'struct', 'union'}
          arg = struct();
          for p = fieldnames(format.type)'
            if isempty(args) return; end
            [arg.(p{1}), args] = self.args2dict(format.type.(p{1}), args);
          end

        case {'array', 'sequence'}
          arg = {};
          if isempty(args) return; end
          a = args{1}; args(1) = [];
          for i = 1:length(a)
            [v, a] = self.args2dict(format.type, a);
            arg(i) = {v};
          end

        case 'optional'
          if length(args{1})
            [arg, args] = self.args2dict(format.type, args);
          else
            arg = args{1}; args(1) = [];
          end

        otherwise
          if iscell(args)
            arg = args{1}; args(1) = [];
          else
            arg = args; args = {};
          end
      end
    end


    % --- meta2doc ---------------------------------------------------------
    %
    function doc = meta2doc(self, input, meta)
      doc = '';
      for p = fieldnames(input)'
        if isfield(meta, p{1}) m = meta.(p{1}); else m = struct(); end
        doc = [doc self.dict2doc('', p{1}, input.(p{1}), m)];
      end
    end


    % --- dict2doc ---------------------------------------------------------
    %
    function prompt = dict2doc(self, indent, name, format, meta)
      prompt = sprintf('\n%s%s', indent, format.kind);
      if name
        prompt = sprintf('%s %s', prompt, name);
      end
      if isfield(format, 'length')
        prompt = sprintf('%s[%d]', prompt, format.length);
      end
      if isfield(meta, 'doc')
        prompt = sprintf('%s: %s', prompt, meta.doc);
      end
      if ~strcmp(format.kind, 'optional')
        if isfield(meta, 'default')
          prompt = strcat(prompt, ' (', meta.default, ')');
        end
      end
      if isfield(meta, 'items') items = meta.items; else items = struct(); end

      switch format.kind
        case {'struct', 'union'}
          for p = fieldnames(format.type)'
            if isfield(items, p) m = items.(p); else m = struct(); end
            prompt = strcat(prompt, self.dict2doc(strcat(indent, '| '), ...
                                                  strcat(name, '.', p), ...
                                                  format.type.(p), m));
          end

        case {'array', 'sequence'}
          prompt = strcat(prompt, self.dict2doc(strcat(indent, '| '), ...
                                                strcat(name, '[i]'), ...
                                                format.type, struct()));
          for i = fieldnames(items)'
            prompt = strcat(prompt, ...
                            self.dict2doc(strcat(indent, '| '), ...
                                          strcat(name, '[', i{1}, ']'), ...
                                          format.type, items.(i{1})));
          end

        case 'optional'
          if isfield(meta, 'default') m = meta.default; else m = struct(); end
          prompt = strcat(prompt, self.dict2doc(strcat(indent, '| '), ...
                                                '', format.type, m))
      end
    end
  end
end
