/*
 * Copyright (c) 2015 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Tue Jun  2 2015
 */
#include <signal.h>
#include <string.h>

#include "mex.h"

#include "libgenomix.h"


/* --- local data ---------------------------------------------------------- */

#define json_skip_whitespace(json) (json) += strspn(json, " \t\n\r")


static mxArray *	j2s_value(const char **json, char *emsg,
				size_t emsgsize);


/* --- genomix.connect ----------------------------------------------------- */

void
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  char emsg[256], tmp[128];
  char *key;
  int s;

  if (nrhs < 1)
    mexErrMsgIdAndTxt("genomix:args", "invalid arguments");

  key = mxArrayToString(prhs[0]);
  if (!key)
    mexErrMsgIdAndTxt("genomix:args", "invalid arguments");

  if (!strcmp(key, "get")) {
    const char *args[nrhs];
    const char *r;
    int id;
    int i;

    if (nrhs < 3)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    id = mxGetScalar(prhs[1]);

    for(i = 2; i < nrhs; i++)
      args[i-2] = mxArrayToString(prhs[i]);
    args[i-2] = NULL;
    if (!args[0])
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");

    r = genomix_get(id, args[0], &args[1], emsg, sizeof(emsg));
    if (!r)
      mexErrMsgIdAndTxt("genomix:run", emsg);

    if (plhs > 0) {
      if (r[0] == '{' || r[0] == '[' || r[0] == '"')
        plhs[0] = j2s_value(&r, emsg, sizeof(emsg));
      else
        plhs[0] = mxCreateString(r);
      if (!plhs[0]) mexErrMsgIdAndTxt("genomix:run", emsg);
    }

    for(i = 2; i < nrhs; i++)
      mxFree((void *)args[i-2]);


  } else if (!strcmp(key, "doevents")) {
    int timeout = 0;

    if (nrhs > 1) {
      double *t;
      if (!mxIsDouble(prhs[1]))
        mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
      t = mxGetPr(prhs[1]);
      if (t) timeout = *t * 1000;
    }
    genomix_doevent(timeout);

  } else if (!strcmp(key, "json")) {
    const char *in, *json;

    if (nrhs != 2)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    in = mxArrayToString(prhs[1]);
    if (!in)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");

    json = in;
    if (nlhs > 0) {
      plhs[0] = j2s_value(&json, emsg, sizeof(emsg));
      if (!plhs[0]) mexErrMsgIdAndTxt("genomix:run", emsg);
    }
    mxFree((void *)in);


  } else if (!strcmp(key, "connect")) {
    char *host = NULL;
    char *session = NULL;
    mxArray *rval;

    if (nrhs > 3)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    if (nrhs > 2) {
      session = mxArrayToString(prhs[2]);
      if (!session)
        mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    }
    if (nrhs > 1) {
      host = mxArrayToString(prhs[1]);
      if (!host)
        mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    }

    s = genomix_connect(host, session, emsg, sizeof(emsg));

    if (host) mxFree(host);
    if (session) mxFree(session);
    if (s < 0) mexErrMsgIdAndTxt("genomix:run", emsg);

    if (!mexIsLocked()) mexLock();
    if (nlhs > 0)
      plhs[0] = mxCreateDoubleScalar(s);

  } else if (!strcmp(key, "disconnect")) {
    int id, s;

    if (nrhs != 2)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    id = mxGetScalar(prhs[1]);

    if (!genomix_disconnect(id))
      mexUnlock();

  } else
    mexErrMsgIdAndTxt("genomix:args", "invalid command");

  if (key) mxFree(key);
}


/* --- genomix_event ------------------------------------------------------- */

void
genomix_event(const char *data)
{
  mxArray *prhs[1];
  char emsg[256];

  prhs[0] = j2s_value(&data, emsg, sizeof(emsg));
  mexCallMATLAB(0, NULL, 1, prhs, "genomix.client.doevents");
  mxDestroyArray(prhs[0]);
}


/* --- j2s_tokerror -------------------------------------------------------- */

/** Append an error message with the current token
 */
static void *
j2s_tokerror(const char *what, const char *json, char *emsg, size_t emsgsize)
{
  char tok[128];
  size_t s;

  s = strcspn(json, " \t\n\r{}[],:\"");
  if (!s) s = 1;
  if (s > sizeof(tok)-1) s = sizeof(tok)-1;
  strncpy(tok, json, s);

  snprintf(emsg, emsgsize, "unexpected %s \"%s\"", what, tok);
  return NULL;
}


/* --- j2s_value ----------------------------------------------------------- */

/** Parse a JSON value
 */
static mxArray *
j2s_value(const char **json, char *emsg, size_t emsgsize)
{
  static const char tok_null[] = "null";
  static const char tok_true[] = "true";
  static const char tok_false[] = "false";

  json_skip_whitespace(*json);
  switch(**json) {
    case 0:
      return NULL;

    case '{': /* object */ {
      mxArray **va = NULL;
      mxArray *k, *v, *s;
      char *key;
      size_t na = 0;
      int i, m, isarray;

      ++(*json);
      json_skip_whitespace(*json);

      s = mxCreateStructArray(1, &(mwSize){1}, 0, NULL);
      if (!s) return s;
      m = isarray = 0;

      while(**json != '}') {
        /* key: string */
        if (**json != '"')
          return j2s_tokerror("non-string key", *json, emsg, emsgsize);
        if (!(k = j2s_value(json, emsg, emsgsize))) return NULL;
        key = mxArrayToString(k);
        if (!key) return NULL;

        if (isarray || (key[0] >= '0' && key[0] <= '9')) {
          isarray = 1;
          i = atoi(key);
          if (i > m) m = i;
        } else
          i = mxAddField(s, key);
        mxFree(key);

        json_skip_whitespace(*json);
        if (**json != ':')
          return j2s_tokerror("element separator", *json, emsg, emsgsize);
        ++(*json);

        /* value */
        if (isarray) {
          if (i >= na) {
            size_t ona = na;
            if (na) na <<= 1; else na = 16;
            va = realloc(va, na * sizeof(*va));
            if (!va) return NULL;
            memset(&va[ona], 0, (char *)&va[na] - (char *)&va[ona]);
          }
          if (!(va[i] = j2s_value(json, emsg, emsgsize))) return NULL;
        } else {
          if (!(v = j2s_value(json, emsg, emsgsize))) return NULL;
          mxSetFieldByNumber(s, 0, i, v);
        }

        json_skip_whitespace(*json);
        if (**json == ',') {
          (*json)++;
          json_skip_whitespace(*json);
        } else if (**json != '}')
          return j2s_tokerror("element", *json, emsg, emsgsize);
      }
      ++(*json);

      if (isarray) {
        mxDestroyArray(s);
        m++;
        s = mxCreateCellArray(1, &(mwSize){m});
        if (s) while(m--) if (va[m]) mxSetCell(s, m, va[m]);
        free(va);
      }

      return s;
    }

    case '[': /* array */ {
      mxArray *a, **v;
      size_t i, n;

      ++(*json);
      json_skip_whitespace(*json);

      i = 0;
      n = 16;
      v = malloc(n * sizeof(*v));
      if (!v) return NULL;
      while(**json != ']') {
        if (i >= n) {
          n <<= 1;
          v = realloc(v, n * sizeof(*v));
          if (!v) return NULL;
        }
        if (!(v[i] = j2s_value(json, emsg, emsgsize))) return NULL;

        json_skip_whitespace(*json);
        if (**json == ',') {
          (*json)++;
          json_skip_whitespace(*json);
        } else if (**json != ']')
          return j2s_tokerror("element", *json, emsg, emsgsize);

        i++;
      }
      ++(*json);

      a = mxCreateCellArray(1, &(mwSize){i});
      if (a) while(i--) mxSetCell(a, i, v[i]);
      free(v);
      return a;
    }

    case '"': /* string */ {
      const char *start = ++(*json);
      int quote = 0;
      char *new, *p;
      mxArray *v;

      while (**json != '"' && **json) {
        if (**json == '\\') {
          if (!*((*json)++)) break;
          if (**json == '"') quote = 1;
        }
        (*json)++;
      }
      if (!**json)
        return j2s_tokerror("end of string", start, emsg, emsgsize);
      (*json)++;

      new = malloc(*json - start);
      if (!new) return NULL;
      strncpy(new, start, *json - start - 1);
      new[*json - start - 1] = 0;

      if (quote) {
        p = new;
        while((p = strstr(p, "\\"))) {
          if (*p == '"' || *p == '\\')
            memmove(p, p+1, strlen(p)/*including final \0*/);
          p++;
        }
      }
      v = mxCreateString(new);
      free(new);
      return v;
    }

    case 't': /* bare word: true, false, null */
    case 'f':
    case 'n':
      if (!strncmp(*json, tok_true, sizeof(tok_true)-1)) {
        *json += sizeof(tok_true)-1;
        return mxCreateLogicalScalar(true);
      }
      if (!strncmp(*json, tok_false, sizeof(tok_false)-1)) {
        *json += sizeof(tok_false)-1;
        return mxCreateLogicalScalar(false);
      }
      if (!strncmp(*json, tok_null, sizeof(tok_null)-1)) {
        *json += sizeof(tok_null)-1;
        return mxCreateStructArray(1, &(mwSize){1}, 0, NULL);
      }

      return j2s_tokerror("value", *json, emsg, emsgsize);

    default: /* number or error */ {
      const char *start = *json;
      char *end;
      double v;

      v = strtod(start, &end);
      if (end == start) return j2s_tokerror("value", start, emsg, emsgsize);
      *json = end;

      return mxCreateDoubleScalar(v);
    }
  }

  snprintf(emsg, emsgsize, "internal error");
  return NULL;
}
